#!/usr/bin/python
# ################################################################## 
# 
# Copyright 2018 Teradata. All rights reserved.
# TERADATA CONFIDENTIAL AND TRADE SECRET
# 
# Primary Owner: Pankaj Purandare (pankajvinod.purandare@teradata.com)
# Secondary Owner: Mounika Kotha (mounika.kotha@teradata.com)
# 
# Version: 1.2
# Function Version: 1.9
# 
# ################################################################## 

from teradataml.common.wrapper_utils import AnalyticsWrapperUtils
from teradataml.common.utils import UtilFuncs
from teradataml.context.context import *
from teradataml.dataframe.dataframe import DataFrame
from teradataml.common.aed_utils import AedUtils
from teradataml.analytics.analytic_query_generator import AnalyticQueryGenerator
from teradataml.common.exceptions import TeradataMlException
from teradataml.common.messages import Messages
from teradataml.common.messagecodes import MessageCodes
from teradataml.common.constants import TeradataConstants
from teradataml.dataframe.dataframe_utils import DataFrameUtils as df_utils
from teradataml.options.display import display
from teradataml.analytics.mle.GLM import GLM

from teradataml.common.utils import package_deprecation

@package_deprecation('16.20.00.01', 'teradataml.analytics.sqle')
class GLMPredict:
    
    def __init__(self,
        modeldata = None,
        newdata = None,
        terms = None,
        family = None,
        linkfunction = "CANONICAL") :
        """
        DESCRIPTION:
            The GLMPredict function uses the model generated by the function GLM 
            to perform generalized linear model prediction on new input data.
        
        
        PARAMETERS:
            modeldata:
                Required Argument.
                Specifies the teradataml DataFrame containing the model data.
            
            newdata:
                Required Argument.
                Specifies the teradataml DataFrame containing the input data.
            
            terms:
                Optional Argument.
                Specifies the names of input teradataml DataFrame columns to copy to 
                the output table.
                Types: str OR list of Strings (str)
            
            family:
                Optional Argument.
                Specifies the distribution exponential family. The default value is 
                read from model table. If you specify this argument, you must give it 
                the same value that you used for the Family argument of the function  
                when you generated the model table.
                Permitted Values: LOGISTIC, BINOMIAL, POISSON, GAUSSIAN, GAMMA, 
                INVERSE_GAUSSIAN, NEGATIVE_BINOMIAL
                Types: str
            
            linkfunction:
                Optional Argument.
                The canonical link functions (default link functions) and the link 
                functions that are allowed for each exponential family Note: Use the 
                same value that you used for the Link argument of the function  when 
                you generated the model table.
                Default Value: "CANONICAL"
                Permitted Values: CANONICAL, IDENTITY, INVERSE, LOG, 
                COMPLEMENTARY_LOG_LOG, SQUARE_ROOT, INVERSE_MU_SQUARED, LOGIT, 
                PROBIT, CAUCHIT
                Types: str
        
        RETURNS:
            Instance of GLMPredict.
            Output teradataml DataFrames can be accessed using attribute
            references, such as GLMPredictObj.<attribute_name>.
            Output teradataml DataFrame attribute name is:
                result
        
        
        RAISES:
            TeradataMlException
        
        
        EXAMPLES:
            # Load the data to run the example.
            load_example_data("GLMPredict", ["admissions_test","admissions_train","housing_test","housing_train"])
            
            # Create teradataml DataFrame objects.
            admissions_test = DataFrame.from_table("admissions_test")
            admissions_train = DataFrame.from_table("admissions_train")
            housing_test = DataFrame.from_table("housing_test")
            housing_train = DataFrame.from_table("housing_train")

            # Example 1 -
            # First train the data, i.e., create a GLM Model
            glm_out = GLM(formula = "admitted ~ stats + masters + gpa + programming",
                         family = "LOGISTIC",
                         linkfunction = "LOGIT",
                         data = admissions_train,
                         weights = "1",
                         threshold = 0.01,
                         maxit = 25,
                         step = False,
                         intercept = True
                         )

            # Run predict on the output of GLM
            glm_predict_out1 = GLMPredict(modeldata = glm_out,
                                          newdata = admissions_test,
                                          terms = ["id","masters","gpa","stats","programming","admitted"],
                                          family = "LOGISTIC",
                                          linkfunction = "LOGIT"
                                          )

            # Print the results.
            glm_predict_out1.result

            # Example 2 -
            # First train the data, i.e., create a GLM Model
            glm_out_hs = GLM(formula = "price  ~ recroom  + lotsize  + stories  + garagepl + gashw \
                                       + bedrooms  + driveway  + airco  + homestyle + bathrms  + fullbase \
                                       + prefarea",
                              family = "GAUSSIAN",
                              linkfunction = "IDENTITY",
                              data = housing_train,
                              weights = "1",
                              threshold = 0.01,
                              maxit = 25,
                              step = False,
                              intercept = True
                              )

            # Run predict on the output of GLM by passing coefficients
            glm_predict_out2 = GLMPredict(modeldata = glm_out_hs.coefficients,
                                          newdata = housing_test,
                                          terms = ["sn", "price"],
                                          family = "GAUSSIAN",
                                          linkfunction = "CANONICAL"
                                          )

            # Print the results.
            glm_predict_out2
        
        """
        self.modeldata  = modeldata 
        self.newdata  = newdata 
        self.terms  = terms 
        self.family  = family 
        self.linkfunction  = linkfunction 
        
        # Create TeradataPyWrapperUtils instance which contains validation functions.
        self.__awu = AnalyticsWrapperUtils()
        self.__aed_utils = AedUtils()
        
        # Create argument information matrix to do parameter checking
        self.__arg_info_matrix = []
        self.__arg_info_matrix.append(["modeldata", self.modeldata, False, "DataFrame"])
        self.__arg_info_matrix.append(["newdata", self.newdata, False, "DataFrame"])
        self.__arg_info_matrix.append(["terms", self.terms, True, "str"])
        self.__arg_info_matrix.append(["family", self.family, True, "str"])
        self.__arg_info_matrix.append(["linkfunction", self.linkfunction, True, "str"])
        
        # Perform the function validations
        self.__validate()
        # Generate the ML query
        self.__form_tdml_query()
        # Execute ML query
        self.__execute()
        
    def __validate(self) :
        """
        Function to validate sqlmr function arguments, which verifies missing 
        arguments, input argument and table types. Also processes the 
        argument values.
        """
        # Make sure that a non-NULL value has been supplied for all mandatory arguments
        self.__awu._validate_missing_required_arguments(self.__arg_info_matrix)
        
        # Make sure that a non-NULL value has been supplied correct type of argument
        self.__awu._validate_argument_types(self.__arg_info_matrix)
        
        # Check to make sure input table types are strings or data frame objects or of valid type.
        self.__awu._validate_input_table_datatype(self.newdata, "newdata", None)
        self.__awu._validate_input_table_datatype(self.modeldata, "modeldata", GLM)
        
        # Check for permitted values
        family_permitted_values = ["LOGISTIC", "BINOMIAL", "POISSON", "GAUSSIAN", "GAMMA", "INVERSE_GAUSSIAN", "NEGATIVE_BINOMIAL"]
        self.__awu._validate_permitted_values(self.family, family_permitted_values, "family")
        
        linkfunction_permitted_values = ["CANONICAL", "IDENTITY", "INVERSE", "LOG", "COMPLEMENTARY_LOG_LOG", "SQUARE_ROOT", "INVERSE_MU_SQUARED", "LOGIT", "PROBIT", "CAUCHIT"]
        self.__awu._validate_permitted_values(self.linkfunction, linkfunction_permitted_values, "linkfunction")
        
        # Check whether the input columns passed to the argument are not empty.
        # Also check whether the input columns passed to the argument valid or not.
        self.__awu._validate_input_columns_not_empty(self.terms, "terms")
        self.__awu._validate_dataframe_has_argument_columns(self.terms, "terms", self.newdata, "newdata")
        
        
    def __form_tdml_query(self) :
        """
        Function to generate the analaytical function queries. The function defines 
        variables and list of arguments required to form the query.
        """
        if isinstance(self.modeldata, GLM):
            self.modeldata = self.modeldata._mlresults[0]
        
        # Output table arguments list
        self.__func_output_args_sql_names = []
        self.__func_output_args = []
        
        # Generate lists for rest of the function arguments
        self.__func_other_arg_sql_names = []
        self.__func_other_args = []
        self.__func_other_arg_json_datatypes = []
        
        if self.terms is not None:
            self.__func_other_arg_sql_names.append("Accumulate")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.terms,"'"))
            self.__func_other_arg_json_datatypes.append("COLUMNS")
        
        if self.family is not None:
            self.__func_other_arg_sql_names.append("Family")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.family,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        if self.linkfunction is not None and self.linkfunction != "CANONICAL":
            self.__func_other_arg_sql_names.append("LinkFunction")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.linkfunction,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        
        # Declare empty lists to hold input table information.
        self.__func_input_arg_sql_names = []
        self.__func_input_table_view_query = []
        self.__func_input_dataframe_type = []
        self.__func_input_distribution = []
        self.__func_input_partition_by_cols = []
        self.__func_input_order_by_cols = []
        
        # Process newdata
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.newdata, False)
        self.__func_input_distribution.append("FACT")
        self.__func_input_arg_sql_names.append("input")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append("ANY")
        self.__func_input_order_by_cols.append("NA_character_")
        
        # Process modeldata
        modeldata_order_columns = ['attribute', 'predictor', 'category', 'estimate']
        modeldata_order_columns = UtilFuncs._teradata_collapse_arglist(modeldata_order_columns, "\"")
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.modeldata, False)
        self.__func_input_distribution.append("DIMENSION")
        self.__func_input_arg_sql_names.append("model")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append("NA_character_")
        self.__func_input_order_by_cols.append(modeldata_order_columns)
        
        function_name = "GLMPredict"
        # Create instance to generate SQLMR.
        aqg_obj = AnalyticQueryGenerator(function_name 
                ,self.__func_input_arg_sql_names 
                ,self.__func_input_table_view_query 
                ,self.__func_input_dataframe_type 
                ,self.__func_input_distribution 
                ,self.__func_input_partition_by_cols 
                ,self.__func_input_order_by_cols 
                ,self.__func_other_arg_sql_names 
                ,self.__func_other_args 
                ,self.__func_other_arg_json_datatypes 
                ,self.__func_output_args_sql_names 
                ,self.__func_output_args)
        # Invoke call to SQL-MR generation.
        self.sqlmr_query = aqg_obj._gen_sqlmr_select_stmt_sql()
        
        # Print SQL-MR query if requested to do so.
        if display.print_sqlmr_query:
            print(self.sqlmr_query)
        
    def __execute(self) :
        """
        Function to execute SQL-MR queries. 
        And creates dataframes for the required SQL-MR outputs.
        """
        # Generate STDOUT table name and add it to the output table list.
        sqlmr_stdout_temp_tablename = UtilFuncs._generate_temp_table_name(prefix = "td_sqlmr_out_", use_default_database = True, gc_on_quit = True, quote=False)
        try:
            UtilFuncs._create_view(sqlmr_stdout_temp_tablename, self.sqlmr_query)
        except Exception as emsg:
            raise TeradataMlException(Messages.get_message(MessageCodes.TDMLDF_EXEC_SQL_FAILED, str(emsg)), MessageCodes.TDMLDF_EXEC_SQL_FAILED)
        
        # Update output table data frames.
        self._mlresults = []
        self.result = self.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(sqlmr_stdout_temp_tablename), source_type="table", database_name=UtilFuncs._extract_db_name(sqlmr_stdout_temp_tablename))
        self._mlresults.append(self.result)
        
    def __repr__(self) :
        """
        Returns the string representation for a GLMPredict class instance.
        """
        repr_string="############ STDOUT Output ############"
        repr_string = "{}\n\n{}".format(repr_string,self.result)
        return repr_string
        
