#!/usr/bin/python
# ################################################################## 
# 
# Copyright 2018 Teradata. All rights reserved.
# TERADATA CONFIDENTIAL AND TRADE SECRET
# 
# Primary Owner: Pankaj Purandare (pankajvinod.purandare@teradata.com)
# Secondary Owner: Mounika Kotha (mounika.kotha@teradata.com)
# 
# Version: 1.2
# Function Version: 1.4
# 
# ################################################################## 

from teradataml.common.wrapper_utils import AnalyticsWrapperUtils
from teradataml.common.utils import UtilFuncs
from teradataml.context.context import *
from teradataml.dataframe.dataframe import DataFrame
from teradataml.common.aed_utils import AedUtils
from teradataml.analytics.analytic_query_generator import AnalyticQueryGenerator
from teradataml.common.exceptions import TeradataMlException
from teradataml.common.messages import Messages
from teradataml.common.messagecodes import MessageCodes
from teradataml.common.constants import TeradataConstants
from teradataml.dataframe.dataframe_utils import DataFrameUtils as df_utils
from teradataml.options.display import display

class SVMSparse:
    
    def __init__(self,
        data = None,
        sample_id_column = None,
        attribute_column = None,
        value_column = None,
        label_column = None,
        cost = 1.0,
        bias = 0.0,
        hash = False,
        hash_buckets = None,
        class_weights = None,
        max_step = 100,
        epsilon = 0.01,
        seed = 0,
        data_sequence_column = None) :
        """
        DESCRIPTION:
            The SVMSparse function takes training data (in sparse format) 
            and outputs a predictive model in binary format, which is input to 
            the functions SVMSparsePredictor and SVMSparseSummary.
        
        
        PARAMETERS:
            data:
                Required Argument.
                Specifies the name of the teradataml DataFrame that contains 
                the training samples.
            
            sample_id_column:
                Required Argument.
                Specifies the name of the name of the column in data, teradataml DataFrame
                that contains the identifiers of the training samples.
                Types: str
            
            attribute_column:
                Required Argument.
                Specifies the name of the name of the column in data, teradataml DataFrame
                that contains the attributes of the samples.
                Types: str
            
            value_column:
                Optional Argument.
                Specifies the name of the name of the column in data, teradataml DataFrame
                that contains the attribute values. By default, each attribute has the value 1.
                Types: str
            
            label_column:
                Required Argument.
                Specifies the name of the name of the column in data, teradataml DataFrame
                that contains the classes of the samples.
                Types: str
            
            cost:
                Optional Argument.
                The regularization parameter in the SVM soft-margin loss function: 
                Must be greater than 0.0. 
                Default Value: 1.0
                Types: float
            
            bias:
                Optional Argument.
                A non-negative value. If the value is greater than zero, each sample 
                x in the training set will be converted to (x, b); that is, it will 
                add another dimension containing the bias value b. This argument 
                addresses situations where not all samples center at 0. 
                Default Value: 0.0
                Types: float
            
            hash:
                Optional Argument.
                Specifies whether to use hash projection on attributes. hash 
                projection can accelerate processing speed but can slightly decrease 
                accuracy. Note: You must use hash projection if the dataset has more 
                features than fit into memory
                Default Value: False
                Types: bool
            
            hash_buckets:
                Optional Argument.
                Valid only if hash is True. Specifies the number of buckets for 
                hash projection. In most cases, the function can determine the 
                appropriate number of buckets from the scale of the input data set. 
                However, if the dataset has a very large number of features, you 
                might have to specify buckets_number to accelerate the function.
                Types: int
            
            class_weights:
                Optional Argument.
                Specifies the weights for different classes. The format is: 
                "classlabel m:weight m, classlabel n:weight n". If weight for a class 
                is given, the cost parameter for this class is weight * cost. A 
                weight larger than 1 often increases the accuracy of the 
                corresponding class; however, it may decrease global accuracy. 
                Classes not assigned a weight in this argument is assigned a weight 
                of 1.0.
                Types: str OR list of Strings (str)
            
            max_step:
                Optional Argument.
                A positive integer value that specifies the maximum number of 
                iterations of the training process. One step means that each sample 
                is seen once by the trainer. The input value must be in the range (0, 
                10000]. 
                Default Value: 100
                Types: int
            
            epsilon:
                Optional Argument.
                Termination criterion. When the difference between the values of the 
                loss function in two sequential iterations is less than this number, 
                the function stops. Must be greater than 0.0. 
                Default Value: 0.01
                Types: float
            
            seed:
                Optional Argument.
                A long integer value used to order the training set randomly and 
                consistently. This value can be used to ensure that the same model 
                will be generated if the function is run multiple times in a given 
                database with the same arguments. The input value must be in the 
                range [0, 9223372036854775807]. The default value is 0.
                Default Value: 0
                Types: int
            
            data_sequence_column:
                Optional Argument.
                Specifies the list of column(s) that uniquely identifies each row of 
                the input argument "data". The argument is used to ensure 
                deterministic results for functions which produce results that vary 
                from run to run.
                Types: str OR list of Strings (str)
        
        RETURNS:
            Instance of SVMSparse.
            Output teradataml dataframes can be accessed using attribute 
            references, such as SVMSparseObj.<attribute_name>.
            Output teradataml dataframe attribute name is: output
        
        
        RAISES:
            TeradataMlException
        
        
        EXAMPLES:
            # Load the data to run the example.
            load_example_data("SVMSparse","svm_iris_input_train")
            
            # Create teradataml DataFrame
            svm_iris_input_train = DataFrame.from_table("svm_iris_input_train")

            # Example 1
            svm_sparse_out = SVMSparse(data=svm_iris_input_train,
                                       sample_id_column='id',
                                       attribute_column='attribute',
                                       label_column='species',
                                       value_column='value1',
                                       max_step=150,
                                       seed=0,
                                       )

            # Print the result DataFrame
            svm_sparse_out.model_table
            svm_sparse_out.output
        
        """
        self.data  = data 
        self.sample_id_column  = sample_id_column 
        self.attribute_column  = attribute_column 
        self.value_column  = value_column 
        self.label_column  = label_column 
        self.cost  = cost 
        self.bias  = bias 
        self.hash  = hash 
        self.hash_buckets  = hash_buckets 
        self.class_weights  = class_weights 
        self.max_step  = max_step 
        self.epsilon  = epsilon 
        self.seed  = seed 
        self.data_sequence_column  = data_sequence_column 
        
        # Create TeradataPyWrapperUtils instance which contains validation functions.
        self.__awu = AnalyticsWrapperUtils()
        self.__aed_utils = AedUtils()
        
        # Create argument information matrix to do parameter checking
        self.__arg_info_matrix = []
        self.__arg_info_matrix.append(["data", self.data, False, "DataFrame"])
        self.__arg_info_matrix.append(["sample_id_column", self.sample_id_column, False, "str"])
        self.__arg_info_matrix.append(["attribute_column", self.attribute_column, False, "str"])
        self.__arg_info_matrix.append(["value_column", self.value_column, True, "str"])
        self.__arg_info_matrix.append(["label_column", self.label_column, False, "str"])
        self.__arg_info_matrix.append(["cost", self.cost, True, "float"])
        self.__arg_info_matrix.append(["bias", self.bias, True, "float"])
        self.__arg_info_matrix.append(["hash", self.hash, True, "bool"])
        self.__arg_info_matrix.append(["hash_buckets", self.hash_buckets, True, "int"])
        self.__arg_info_matrix.append(["class_weights", self.class_weights, True, "str"])
        self.__arg_info_matrix.append(["max_step", self.max_step, True, "int"])
        self.__arg_info_matrix.append(["epsilon", self.epsilon, True, "float"])
        self.__arg_info_matrix.append(["seed", self.seed, True, "int"])
        self.__arg_info_matrix.append(["data_sequence_column", self.data_sequence_column, True, "str"])
        
        # Perform the function validations
        self.__validate()
        # Generate the ML query
        self.__form_tdml_query()
        # Execute ML query
        self.__execute()
        
    def __validate(self) :
        """
        Function to validate sqlmr function arguments, which verifies missing 
        arguments, input argument and table types. Also processes the 
        argument values.
        """
        # Make sure that a non-NULL value has been supplied for all mandatory arguments
        self.__awu._validate_missing_required_arguments(self.__arg_info_matrix)
        
        # Make sure that a non-NULL value has been supplied correct type of argument
        self.__awu._validate_argument_types(self.__arg_info_matrix)
        
        # Check to make sure input table types are strings or data frame objects or of valid type.
        self.__awu._validate_input_table_datatype(self.data, "data", None)
        
        # Check whether the input columns passed to the argument are not empty.
        # Also check whether the input columns passed to the argument valid or not.
        self.__awu._validate_input_columns_not_empty(self.sample_id_column, "sample_id_column")
        self.__awu._validate_dataframe_has_argument_columns(self.sample_id_column, "sample_id_column", self.data, "data")
        
        self.__awu._validate_input_columns_not_empty(self.attribute_column, "attribute_column")
        self.__awu._validate_dataframe_has_argument_columns(self.attribute_column, "attribute_column", self.data, "data")
        
        self.__awu._validate_input_columns_not_empty(self.label_column, "label_column")
        self.__awu._validate_dataframe_has_argument_columns(self.label_column, "label_column", self.data, "data")
        
        self.__awu._validate_input_columns_not_empty(self.value_column, "value_column")
        self.__awu._validate_dataframe_has_argument_columns(self.value_column, "value_column", self.data, "data")
        
        self.__awu._validate_input_columns_not_empty(self.data_sequence_column, "data_sequence_column")
        self.__awu._validate_dataframe_has_argument_columns(self.data_sequence_column, "data_sequence_column", self.data, "data")
        
        
    def __form_tdml_query(self) :
        """
        Function to generate the analytical function queries. The function defines 
        variables and list of arguments required to form the query.
        """
        # Generate temp table names for output table parameters if any.
        self.__model_table_temp_tablename = UtilFuncs._generate_temp_table_name(prefix = "td_svmsparse0", use_default_database = True, gc_on_quit = True, quote=False, table_type = TeradataConstants.TERADATA_TABLE)
        
        # Output table arguments list
        self.__func_output_args_sql_names = ["ModelTable"]
        self.__func_output_args = [self.__model_table_temp_tablename]
        
        # Generate lists for rest of the function arguments
        self.__func_other_arg_sql_names = []
        self.__func_other_args = []
        self.__func_other_arg_json_datatypes = []
        
        self.__func_other_arg_sql_names.append("IDColumn")
        self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(UtilFuncs._teradata_quote_arg(self.sample_id_column,"\""),"'"))
        self.__func_other_arg_json_datatypes.append("COLUMN_NAMES")
        
        self.__func_other_arg_sql_names.append("AttributeNameColumn")
        self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(UtilFuncs._teradata_quote_arg(self.attribute_column,"\""),"'"))
        self.__func_other_arg_json_datatypes.append("COLUMN_NAMES")
        
        self.__func_other_arg_sql_names.append("ResponseColumn")
        self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(UtilFuncs._teradata_quote_arg(self.label_column,"\""),"'"))
        self.__func_other_arg_json_datatypes.append("COLUMN_NAMES")
        
        if self.value_column is not None:
            self.__func_other_arg_sql_names.append("AttributeValueColumn")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(UtilFuncs._teradata_quote_arg(self.value_column,"\""),"'"))
            self.__func_other_arg_json_datatypes.append("COLUMN_NAMES")
        
        if self.hash is not None and self.hash != False:
            self.__func_other_arg_sql_names.append("HashProjection")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.hash,"'"))
            self.__func_other_arg_json_datatypes.append("BOOLEAN")
        
        if self.hash_buckets is not None:
            self.__func_other_arg_sql_names.append("HashBuckets")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.hash_buckets,"'"))
            self.__func_other_arg_json_datatypes.append("INTEGER")
        
        if self.cost is not None and self.cost != 1:
            self.__func_other_arg_sql_names.append("Cost")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.cost,"'"))
            self.__func_other_arg_json_datatypes.append("DOUBLE")
        
        if self.bias is not None and self.bias != 0:
            self.__func_other_arg_sql_names.append("Bias")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.bias,"'"))
            self.__func_other_arg_json_datatypes.append("DOUBLE")
        
        if self.class_weights is not None:
            self.__func_other_arg_sql_names.append("ClassWeights")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.class_weights,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        if self.max_step is not None and self.max_step != 100:
            self.__func_other_arg_sql_names.append("MaxStep")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.max_step,"'"))
            self.__func_other_arg_json_datatypes.append("INTEGER")
        
        if self.epsilon is not None and self.epsilon != 0.01:
            self.__func_other_arg_sql_names.append("Epsilon")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.epsilon,"'"))
            self.__func_other_arg_json_datatypes.append("DOUBLE")
        
        if self.seed is not None and self.seed != 0:
            self.__func_other_arg_sql_names.append("Seed")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.seed,"'"))
            self.__func_other_arg_json_datatypes.append("LONG")
        
        # Generate lists for rest of the function arguments
        sequence_input_by_list = []
        if self.data_sequence_column is not None:
            sequence_input_by_list.append("InputTable:" + UtilFuncs._teradata_collapse_arglist(self.data_sequence_column,""))
        
        if len(sequence_input_by_list) > 0:
            self.__func_other_arg_sql_names.append("SequenceInputBy")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(sequence_input_by_list,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        
        # Declare empty lists to hold input table information.
        self.__func_input_arg_sql_names = []
        self.__func_input_table_view_query = []
        self.__func_input_dataframe_type = []
        self.__func_input_distribution = []
        self.__func_input_partition_by_cols = []
        self.__func_input_order_by_cols = []
        
        # Process data
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.data, False)
        self.__func_input_distribution.append("NONE")
        self.__func_input_arg_sql_names.append("InputTable")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append("NA_character_")
        self.__func_input_order_by_cols.append("NA_character_")
        
        function_name = "SVMSparse"
        # Create instance to generate SQLMR.
        aqg_obj = AnalyticQueryGenerator(function_name 
                ,self.__func_input_arg_sql_names 
                ,self.__func_input_table_view_query 
                ,self.__func_input_dataframe_type 
                ,self.__func_input_distribution 
                ,self.__func_input_partition_by_cols 
                ,self.__func_input_order_by_cols 
                ,self.__func_other_arg_sql_names 
                ,self.__func_other_args 
                ,self.__func_other_arg_json_datatypes 
                ,self.__func_output_args_sql_names 
                ,self.__func_output_args)
        # Invoke call to SQL-MR generation.
        self.sqlmr_query = aqg_obj._gen_sqlmr_select_stmt_sql()
        
        # Print SQL-MR query if requested to do so.
        if display.print_sqlmr_query:
            print(self.sqlmr_query)
        
    def __execute(self) :
        """
        Function to execute SQL-MR queries. 
        And create dataframes for the required SQL-MR outputs.
        """
        # Generate STDOUT table name and add it to the output table list.
        sqlmr_stdout_temp_tablename = UtilFuncs._generate_temp_table_name(prefix = "td_sqlmr_out_", use_default_database = True, gc_on_quit = True, quote=False, table_type = TeradataConstants.TERADATA_TABLE)
        try:
            UtilFuncs._create_table(sqlmr_stdout_temp_tablename, self.sqlmr_query)
        except Exception as emsg:
            raise TeradataMlException(Messages.get_message(MessageCodes.TDMLDF_EXEC_SQL_FAILED, str(emsg)), MessageCodes.TDMLDF_EXEC_SQL_FAILED)
        
        # Update output table data frames.
        self._mlresults = []
        self.model_table = self.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(self.__model_table_temp_tablename), source_type="table", database_name=UtilFuncs._extract_db_name(self.__model_table_temp_tablename))
        self.output = self.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(sqlmr_stdout_temp_tablename), source_type="table", database_name=UtilFuncs._extract_db_name(sqlmr_stdout_temp_tablename))
        self._mlresults.append(self.model_table)
        self._mlresults.append(self.output)
        
    def __repr__(self) :
        """
        Returns the string representation for a SVMSparse class instance.
        """
        repr_string="############ STDOUT Output ############"
        repr_string = "{}\n\n{}".format(repr_string,self.output)
        repr_string="{}\n\n\n############ model_table Output ############".format(repr_string)
        repr_string = "{}\n\n{}".format(repr_string,self.model_table)
        return repr_string
        
