#!/usr/bin/python
# ################################################################## 
# 
# Copyright 2018 Teradata. All rights reserved.
# TERADATA CONFIDENTIAL AND TRADE SECRET
# 
# Primary Owner: Pankaj Purandare (pankajvinod.purandare@teradata.com)
# Secondary Owner: Mounika Kotha (mounika.kotha@teradata.com)
# 
# Version: 1.2
# Function Version: 1.2
# 
# ################################################################## 

from teradataml.common.wrapper_utils import AnalyticsWrapperUtils
from teradataml.common.utils import UtilFuncs
from teradataml.context.context import *
from teradataml.dataframe.dataframe import DataFrame
from teradataml.common.aed_utils import AedUtils
from teradataml.analytics.analytic_query_generator import AnalyticQueryGenerator
from teradataml.common.exceptions import TeradataMlException
from teradataml.common.messages import Messages
from teradataml.common.messagecodes import MessageCodes
from teradataml.common.constants import TeradataConstants
from teradataml.dataframe.dataframe_utils import DataFrameUtils as df_utils
from teradataml.options.display import display

class StringSimilarity:
    
    def __init__(self,
        data = None,
        comparison_columns = None,
        case_sensitive = None,
        accumulate = None,
        data_sequence_column = None) :
        """
        DESCRIPTION:
            The StringSimilarity function calculates the similarity between two
            strings, using either the Jaro, Jaro-Winkler, N-Gram, or Levenshtein 
            distance. The similarity is a value in the range [0, 1].
        
        
        PARAMETERS:
            data:
                Required Argument.
                The teradataml DataFrame contains the string pairs to be compared.
            
            comparison_columns:
                Required Argument.
                Specifies pairs of input teradataml DataFrame columns that contain 
                strings to be compared (column1 and column2), how to compare them 
                (comparison_type), and (optionally) a constant and the name of the 
                output column for their similarity (output_column). The similarity is 
                a value in the range [0, 1]. 
                For comparison_type, use one of these values:
                    •  "jaro": Jaro distance  
                    •  "jaro_winkler": Jaro-Winkler distance (1 for an exact match, 0 otherwise)
                    •  "n-gram": N-gram similarity, if you specify this comparison type, you can specify the 
                       value of N with constant.  
                    •  "LD": Levenshtein distance (the number of edits needed to 
                       transform one string into the other, where edits include 
                       insertions, deletions, or substitutions of individual 
                       characters).
                You can specify a different comparison_type for every pair of 
                columns. The default output_column is "sim_i", where i is the 
                sequence number of the column pair.
                Types: str OR list of Strings (str)
            
            case_sensitive:
                Optional Argument.
                Specifies whether string comparison is case-sensitive. The default 
                value is "false". You can specify either one value for all pairs or 
                one value for each pair. If you specify one value for each pair, then 
                the ith value applies to the ith pair.
                Types: bool OR list of bools
            
            accumulate:
                Optional Argument.
                Specifies the names of input teradataml DataFrame columns to be 
                copied to the output table.
                Types: str OR list of Strings (str)
            
            data_sequence_column:
                Optional Argument.
                Specifies the list of column(s) that uniquely identifies each row of 
                the input argument "data". The argument is used to ensure 
                deterministic results for functions which produce results that vary 
                from run to run.
                Types: str OR list of Strings (str)
        
        RETURNS:
            Instance of StringSimilarity.
            Output teradataml DataFrames can be accessed using attribute 
            references, such as StringSimilarityObj.<attribute_name>.
            Output teradataml DataFrame attribute name is:
                result
        
        
        RAISES:
            TeradataMlException
        
        
        EXAMPLES:
            # Load example data.
            load_example_data("stringsimilarity", "strsimilarity_input")
            
            # Create teradataml DataFrame objects.
            strsimilarity_input = DataFrame.from_table("strsimilarity_input")
            
            # Example 1 - Using "jaro" comparison type with a default output column
            stringsimilarity_out1 = StringSimilarity(data=strsimilarity_input,
                                                     comparison_columns=['jaro (src_text1 , tar_text ) AS jaro1_sim','LD (src_text1 , tar_text, 2) AS ld1_sim','n_gram (src_text1 , tar_text, 2) AS ngram1_sim','jaro_winkler (src_text1 , tar_text, 2) AS jw1_sim'],
                                                     case_sensitive=True,
                                                     accumulate = ["id","src_text1","tar_text"],
                                                     data_sequence_column='id')
            # Print result dataframe.
            stringsimilarity_out1.result
            
            # Example 2 - Using multiple comparison types and with custom output columns
            stringsimilarity_out2 = StringSimilarity(data=strsimilarity_input,
                                                     comparison_columns=['jaro (src_text2 , tar_text ) AS jaro2_sim',
                                                                         'LD (src_text2 , tar_text, 2) AS ld2_sim',
                                                                         'n_gram (src_text2 , tar_text, 2) AS ngram2_sim',
                                                                         'jaro_winkler (src_text2 , tar_text, 2) AS jw2_sim'],
                                                     case_sensitive=True,
                                                     accumulate = ["id","src_text2","tar_text"],
                                                     data_sequence_column='id')
            
            # Print result dataframe.
            stringsimilarity_out2.result
            
            # Example 3- Using a vector for case_sensitive comparisons.
            # Note: The length of the case_sensitive vector must match the 
            # comparison_columns vector argument.
            stringsimilarity_out3 = StringSimilarity(data=strsimilarity_input,
                                                     comparison_columns=["jaro (src_text2, tar_text) AS jaro2_case_sim",
                                                                         "jaro (src_text2, tar_text) AS jaro2_nocase_sim"],
                                                     case_sensitive=[True,False],
                                                     accumulate = ["id","src_text2","tar_text"],
                                                     )
            
            # Print result dataframe.
            print(stringsimilarity_out3)
            
        """
        self.data  = data 
        self.comparison_columns  = comparison_columns 
        self.case_sensitive  = case_sensitive 
        self.accumulate  = accumulate 
        self.data_sequence_column  = data_sequence_column 
        
        # Create TeradataPyWrapperUtils instance which contains validation functions.
        self.__awu = AnalyticsWrapperUtils()
        self.__aed_utils = AedUtils()
        
        # Create argument information matrix to do parameter checking
        self.__arg_info_matrix = []
        self.__arg_info_matrix.append(["data", self.data, False, "DataFrame"])
        self.__arg_info_matrix.append(["comparison_columns", self.comparison_columns, False, "str"])
        self.__arg_info_matrix.append(["case_sensitive", self.case_sensitive, True, "bool"])
        self.__arg_info_matrix.append(["accumulate", self.accumulate, True, "str"])
        self.__arg_info_matrix.append(["data_sequence_column", self.data_sequence_column, True, "str"])
        
        # Perform the function validations
        self.__validate()
        # Generate the ML query
        self.__form_tdml_query()
        # Execute ML query
        self.__execute()
        
    def __validate(self) :
        """
        Function to validate sqlmr function arguments, which verifies missing 
        arguments, input argument and table types. Also processes the 
        argument values.
        """
        
        # Make sure that a non-NULL value has been supplied for all mandatory arguments
        self.__awu._validate_missing_required_arguments(self.__arg_info_matrix)
        
        # Make sure that a non-NULL value has been supplied correct type of argument
        self.__awu._validate_argument_types(self.__arg_info_matrix)
        
        # Check to make sure input table types are strings or data frame objects or of valid type.
        self.__awu._validate_input_table_datatype(self.data, "data", None)
        
        # Check whether the input columns passed to the argument are not empty.
        # Also check whether the input columns passed to the argument valid or not.
        self.__awu._validate_input_columns_not_empty(self.accumulate, "accumulate")
        self.__awu._validate_dataframe_has_argument_columns(self.accumulate, "accumulate", self.data, "data")
        
        self.__awu._validate_input_columns_not_empty(self.data_sequence_column, "data_sequence_column")
        self.__awu._validate_dataframe_has_argument_columns(self.data_sequence_column, "data_sequence_column", self.data, "data")
        
        
    def __form_tdml_query(self) :
        """
        Function to generate the analytical function queries. The function defines 
        variables and list of arguments required to form the query.
        """
        
        # Output table arguments list
        self.__func_output_args_sql_names = []
        self.__func_output_args = []
        
        # Generate lists for rest of the function arguments
        self.__func_other_arg_sql_names = []
        self.__func_other_args = []
        self.__func_other_arg_json_datatypes = []
        
        if self.accumulate is not None:
            self.__func_other_arg_sql_names.append("Accumulate")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(UtilFuncs._teradata_quote_arg(self.accumulate,"\""),"'"))
            self.__func_other_arg_json_datatypes.append("COLUMNS")
        
        self.__func_other_arg_sql_names.append("ComparisonColumnPairs")
        self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.comparison_columns,"'"))
        self.__func_other_arg_json_datatypes.append("STRING")
        
        if self.case_sensitive is not None:
            self.__func_other_arg_sql_names.append("Casesensitive")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.case_sensitive,"'"))
            self.__func_other_arg_json_datatypes.append("BOOLEAN")
        
        # Generate lists for rest of the function arguments
        sequence_input_by_list = []
        if self.data_sequence_column is not None:
            sequence_input_by_list.append("input:" + UtilFuncs._teradata_collapse_arglist(self.data_sequence_column,""))
        
        if len(sequence_input_by_list) > 0:
            self.__func_other_arg_sql_names.append("SequenceInputBy")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(sequence_input_by_list,"'"))
            self.__func_other_arg_json_datatypes.append("STRING")
        
        
        # Declare empty lists to hold input table information.
        self.__func_input_arg_sql_names = []
        self.__func_input_table_view_query = []
        self.__func_input_dataframe_type = []
        self.__func_input_distribution = []
        self.__func_input_partition_by_cols = []
        self.__func_input_order_by_cols = []
        
        # Process data
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.data, False)
        self.__func_input_distribution.append("FACT")
        self.__func_input_arg_sql_names.append("input")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append("ANY")
        self.__func_input_order_by_cols.append("NA_character_")
        
        function_name = "StringSimilarity"
        # Create instance to generate SQLMR.
        aqg_obj = AnalyticQueryGenerator(function_name 
                ,self.__func_input_arg_sql_names 
                ,self.__func_input_table_view_query 
                ,self.__func_input_dataframe_type 
                ,self.__func_input_distribution 
                ,self.__func_input_partition_by_cols 
                ,self.__func_input_order_by_cols 
                ,self.__func_other_arg_sql_names 
                ,self.__func_other_args 
                ,self.__func_other_arg_json_datatypes 
                ,self.__func_output_args_sql_names 
                ,self.__func_output_args)
        # Invoke call to SQL-MR generation.
        self.sqlmr_query = aqg_obj._gen_sqlmr_select_stmt_sql()
        
        # Print SQL-MR query if requested to do so.
        if display.print_sqlmr_query:
            print(self.sqlmr_query)
        
    def __execute(self) :
        """
        Function to execute SQL-MR queries. 
        Create DataFrames for the required SQL-MR outputs.
        """
        # Generate STDOUT table name and add it to the output table list.
        sqlmr_stdout_temp_tablename = UtilFuncs._generate_temp_table_name(prefix = "td_sqlmr_out_", use_default_database = True, gc_on_quit = True, quote=False)
        try:
            UtilFuncs._create_view(sqlmr_stdout_temp_tablename, self.sqlmr_query)
        except Exception as emsg:
            raise TeradataMlException(Messages.get_message(MessageCodes.TDMLDF_EXEC_SQL_FAILED, str(emsg)), MessageCodes.TDMLDF_EXEC_SQL_FAILED)
        
        # Update output table data frames.
        self._mlresults = []
        self.result = self.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(sqlmr_stdout_temp_tablename), source_type="table", database_name=UtilFuncs._extract_db_name(sqlmr_stdout_temp_tablename))
        self._mlresults.append(self.result)
        
    def __repr__(self) :
        """
        Returns the string representation for a StringSimilarity class instance.
        """
        repr_string="############ STDOUT Output ############"
        repr_string = "{}\n\n{}".format(repr_string,self.result)
        return repr_string
        
