#!/usr/bin/python
# ################################################################## 
# 
# Copyright 2018 Teradata. All rights reserved.
# TERADATA CONFIDENTIAL AND TRADE SECRET
# 
# Primary Owner: Pankaj Purandare (pankajvinod.purandare@teradata.com)
# Secondary Owner: Mounika Kotha (mounika.kotha@teradata.com)
# 
# Version: 1.1
# Function Version: 1.0
# 
# ################################################################## 

from teradataml.common.wrapper_utils import AnalyticsWrapperUtils
from teradataml.common.utils import UtilFuncs
from teradataml.context.context import *
from teradataml.dataframe.dataframe import DataFrame
from teradataml.common.aed_utils import AedUtils
from teradataml.analytics.analytic_query_generator import AnalyticQueryGenerator
from teradataml.common.exceptions import TeradataMlException
from teradataml.common.messages import Messages
from teradataml.common.messagecodes import MessageCodes
from teradataml.common.constants import TeradataConstants
from teradataml.dataframe.dataframe_utils import DataFrameUtils as df_utils
from teradataml.options.display import display
from teradataml.analytics.mle.NaiveBayes import NaiveBayes

class NaiveBayesPredict:
    
    def __init__(self,
        modeldata = None,
        newdata = None,
        id_col = None,
        responses = None,
        formula = None) :
        """
        DESCRIPTION:
            The NaiveBayesPredict uses the model output generated by the NaiveBayes
            to predict the outcomes for a test set of data.
        
        
        PARAMETERS:
            modeldata:
                Required Argument.
                This teradataml DataFrame contains the model data. This argument can accept
                teradataml DataFrame or instance of NaiveBayes class.
            
            newdata:
                Required Argument.
                This teradataml DataFrame defines the input test data.
            
            id_col:
                Required Argument.
                Specifies the name of the column that contains the ID that uniquely
                identifies the test input data.
                Types: str
            
            responses:
                Required Argument.
                Specifies a list of Responses to output.
                Types: str OR list of Strings (str)

            formula:
                Optional Argument. Required when the argument "modeldata" is a 'teradataml DataFrame'.
                A string consisting of "formula". Specifies the model to be fitted. Only
                basic formula of the "col1 ~ col2 + col3 +..." form is supported and
                all variables must be from the same virtual data frame object. The
                response should be column of type real, numeric, integer or boolean.
                Types: str
        
        RETURNS:
            Instance of NaiveBayesPredict.
            Output teradataml DataFrames can be accessed using attribute
            references, such as NaiveBayesPredictObj.<attribute_name>.
            Output teradataml DataFrame attribute name is:
                result
        
        
        RAISES:
            TeradataMlException
        
        
        EXAMPLES:
            # Load the data to run the example
            load_example_data("NaiveBayesPredict",["nb_iris_input_test","nb_iris_input_train"])
            
            # Create teradataml DataFrame objects.
            nb_iris_input_train = DataFrame.from_table("nb_iris_input_train")
            nb_iris_input_test = DataFrame.from_table("nb_iris_input_test")

            # Example 1 -
            # Run the train function
            naivebayes_train = NaiveBayes(formula="species ~ petal_length + sepal_width + petal_width + sepal_length",
                                          data=nb_iris_input_train)

            # Generate prediction using output of train function
            naivebayes_predict_result = NaiveBayesPredict(newdata=nb_iris_input_test,
                                                          modeldata = naivebayes_train,
                                                          id_col = "id",
                                                          responses = ["virginica","setosa","versicolor"]
                                                          )
            
            #Print result dataframe
            naivebayes_predict_result.result
        
        """
        self.formula  = formula 
        self.modeldata  = modeldata 
        self.newdata  = newdata 
        self.id_col  = id_col 
        self.responses  = responses 
        
        # Create TeradataPyWrapperUtils instance which contains validation functions.
        self.__awu = AnalyticsWrapperUtils()
        self.__aed_utils = AedUtils()
        
        # Create argument information matrix to do parameter checking
        self.__arg_info_matrix = []
        self.__arg_info_matrix.append(["formula", self.formula, True, "formula"])
        self.__arg_info_matrix.append(["modeldata", self.modeldata, False, "DataFrame"])
        self.__arg_info_matrix.append(["newdata", self.newdata, False, "DataFrame"])
        self.__arg_info_matrix.append(["id_col", self.id_col, False, "str"])
        self.__arg_info_matrix.append(["responses", self.responses, False, "str"])
        
        # Perform the function validations
        self.__validate()
        # Generate the ML query
        self.__form_tdml_query()
        # Execute ML query
        self.__execute()
        
    def __validate(self) :
        """
        Function to validate sqlmr function arguments, which verifies missing
        arguments, input argument and table types. Also processes the 
        argument values.
        """
        # Make sure that a non-NULL value has been supplied for all mandatory arguments
        self.__awu._validate_missing_required_arguments(self.__arg_info_matrix)
        
        # Make sure that a non-NULL value has been supplied correct type of argument
        self.__awu._validate_argument_types(self.__arg_info_matrix)
        
        # Check to make sure input table types are strings or data frame objects or of valid type.
        self.__awu._validate_input_table_datatype(self.newdata, "newdata", None)
        self.__awu._validate_input_table_datatype(self.modeldata, "modeldata", NaiveBayes)
        
        # Check whether the input columns passed to the argument are not empty.
        # Also check whether the input columns passed to the argument valid or not.
        self.__awu._validate_input_columns_not_empty(self.id_col, "id_col")
        self.__awu._validate_dataframe_has_argument_columns(self.id_col, "id_col", self.newdata, "newdata")
        
        
    def __form_tdml_query(self) :
        """
        Function to generate the analytical function queries. The function defines
        variables and list of arguments required to form the query.
        """
        if isinstance(self.modeldata, NaiveBayes):
            self.formula = self.modeldata.formula
            self.modeldata = self.modeldata._mlresults[0]
        else:
            if self.formula is None:
                raise TeradataMlException(Messages.get_message(MessageCodes.MISSING_ARGS, "formula"),
                                          MessageCodes.MISSING_ARGS)
        
        # Output table arguments list
        self.__func_output_args_sql_names = []
        self.__func_output_args = []
        
        # Generate lists for rest of the function arguments
        self.__func_other_arg_sql_names = []
        self.__func_other_args = []
        self.__func_other_arg_json_datatypes = []
        
        self.__func_other_arg_sql_names.append("IdCol")
        self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.id_col,"'"))
        self.__func_other_arg_json_datatypes.append("COLUMNS")
        
        self.__func_other_arg_sql_names.append("Responses")
        self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(self.responses,"'"))
        self.__func_other_arg_json_datatypes.append("STRING")
        
        # Let's process formula argument
        formula_object = self.__awu._validate_formula_notation(self.formula, self.newdata, "formula")
        # numerical input columns
        numerical_inputs = self.__awu._get_columns_by_type(formula_object,self.newdata, "numerical")
        if len(numerical_inputs) > 0:
            self.__func_other_arg_sql_names.append("NumericInputs")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(numerical_inputs,"'"))
            self.__func_other_arg_json_datatypes.append("COLUMN_NAMES")
        
        # categorical input columns
        categorical_inputs = self.__awu._get_columns_by_type(formula_object,self.newdata, "categorical")
        if len(categorical_inputs) > 0:
            self.__func_other_arg_sql_names.append("CategoricalInputs")
            self.__func_other_args.append(UtilFuncs._teradata_collapse_arglist(categorical_inputs,"'"))
            self.__func_other_arg_json_datatypes.append("COLUMN_NAMES")
        
        
        # Declare empty lists to hold input table information.
        self.__func_input_arg_sql_names = []
        self.__func_input_table_view_query = []
        self.__func_input_dataframe_type = []
        self.__func_input_distribution = []
        self.__func_input_partition_by_cols = []
        self.__func_input_order_by_cols = []
        
        # Process newdata
        self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.newdata, False)
        self.__func_input_distribution.append("FACT")
        self.__func_input_arg_sql_names.append("input")
        self.__func_input_table_view_query.append(self.__table_ref["ref"])
        self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
        self.__func_input_partition_by_cols.append("ANY")
        self.__func_input_order_by_cols.append("NA_character_")
        
        # Process modeldata
        if self.modeldata is not None:
            self.__table_ref = self.__awu._teradata_on_clause_from_dataframe(self.modeldata, False)
            self.__func_input_distribution.append("DIMENSION")
            self.__func_input_arg_sql_names.append("model")
            self.__func_input_table_view_query.append(self.__table_ref["ref"])
            self.__func_input_dataframe_type.append(self.__table_ref["ref_type"])
            self.__func_input_partition_by_cols.append("NA_character_")
            self.__func_input_order_by_cols.append("NA_character_")
        
        function_name = "NaiveBayesPredict"
        # Create instance to generate SQLMR.
        aqg_obj = AnalyticQueryGenerator(function_name 
                ,self.__func_input_arg_sql_names 
                ,self.__func_input_table_view_query 
                ,self.__func_input_dataframe_type 
                ,self.__func_input_distribution 
                ,self.__func_input_partition_by_cols 
                ,self.__func_input_order_by_cols 
                ,self.__func_other_arg_sql_names 
                ,self.__func_other_args 
                ,self.__func_other_arg_json_datatypes 
                ,self.__func_output_args_sql_names 
                ,self.__func_output_args)
        # Invoke call to SQL-MR generation.
        self.sqlmr_query = aqg_obj._gen_sqlmr_select_stmt_sql()
        
        # Print SQL-MR query if requested to do so.
        if display.print_sqlmr_query:
            print(self.sqlmr_query)
        
    def __execute(self) :
        """
        Function to execute SQL-MR queries. 
        And creates dataframes for the required SQL-MR outputs.
        """
        # Generate STDOUT table name and add it to the output table list.
        sqlmr_stdout_temp_tablename = UtilFuncs._generate_temp_table_name(prefix = "td_sqlmr_out_", use_default_database = True, gc_on_quit = True, quote=False)
        try:
            UtilFuncs._create_view(sqlmr_stdout_temp_tablename, self.sqlmr_query)
        except Exception as emsg:
            raise TeradataMlException(Messages.get_message(MessageCodes.TDMLDF_EXEC_SQL_FAILED, str(emsg)), MessageCodes.TDMLDF_EXEC_SQL_FAILED)
        
        # Update output table data frames.
        self._mlresults = []
        self.result = self.__awu._create_data_set_object(df_input=UtilFuncs._extract_table_name(sqlmr_stdout_temp_tablename), source_type="table", database_name=UtilFuncs._extract_db_name(sqlmr_stdout_temp_tablename))
        self._mlresults.append(self.result)
        
    def __repr__(self) :
        """
        Returns the string representation for a NaiveBayesPredict class instance.
        """
        repr_string="############ STDOUT Output ############"
        repr_string = "{}\n\n{}".format(repr_string,self.result)
        return repr_string
        
